import ast
import os
from typing import Union

import pytest

from magniv.build.build import (
    build,
    get_decorated_nodes,
    get_magniv_tasks,
    get_task_files,
    get_task_list,
    save_tasks,
)

TEST_FILE = """from datetime import datetime

from magniv.core import task


@task(schedule="@hourly", description="Hello world, printing the time", key="goodbye_world_1")
def hello_world():
    print(f"Hello world, the time is {datetime.now()}")

@task(schedule="@daily", description="Hello world, printing a second time", key="goodbye_world_2")
def hello_world_2():
    print(f"Hello world, the time is {datetime.now()}")

@task(schedule="@daily", resources={"cpu": "500mi", "memory": "3Gi"})
def resourceful_valid():
    print("using custom resources!")

@task(schedule="@daily", webhook_trigger=True)
def playing_webhooky():
    print("I'M TRIGGERED (by a webhook)!")

@task(schedule="@daily", on_success=['task_b', 'c'])
def task_a():
    print("I")

@task()
def task_b():
    print("i'm only triggered by task A")

@task(key="c", on_success=['task_b'])
def task_c():
    print("i'm only triggered by task A")

def dummy_task():
    print("This is a dummy task")
    return True

if __name__ == "__main__":
    hello_world()
"""


class TestBuild:
    @pytest.fixture
    def file(self, tmpdir):
        tmpdir.mkdir("tasks").join("main.py").write(TEST_FILE)
        tmpdir.join("tasks/requirements.txt").write("magniv")
        return str(tmpdir.join("tasks"))

    def _extracted_from_test_get_decorated_nodes(self, decorated_nodes, arg1, arg2, arg3, arg4):
        assert decorated_nodes[arg1].name == arg2
        assert decorated_nodes[arg1].decorator_list[0].func.id == "task"
        assert decorated_nodes[arg1].decorator_list[0].keywords[0].arg == "schedule"
        assert decorated_nodes[arg1].decorator_list[0].keywords[0].value.s == arg3
        assert decorated_nodes[arg1].decorator_list[0].keywords[2].value.s == arg4

    def get_ast(self, file) -> Union[ast.AST, str]:
        """
        It opens a file, reads it, and parses it into an AST
        :return: The parsed ast and the filepath
        """
        with open(f"{file}/main.py") as f:
            parsed_ast = ast.parse(f.read())
        return parsed_ast, file

    def test_get_decorated_nodes(self, file):
        """
        `get_decorated_nodes` returns a list of all the decorated functions in a Python file
        """
        parsed_ast, filepath = self.get_ast(file)
        decorated_nodes, _ = get_decorated_nodes(parsed_ast)
        self._extracted_from_test_get_decorated_nodes(
            decorated_nodes, 0, "hello_world", "@hourly", "goodbye_world_1"
        )
        self._extracted_from_test_get_decorated_nodes(
            decorated_nodes, 1, "hello_world_2", "@daily", "goodbye_world_2"
        )

    def test_get_magniv_tasks(self, file):
        """
        It takes a filepath, parses the file, finds all the decorated functions, and returns a list of
        dictionaries containing the information about each task
        """
        parsed_ast, filepath = self.get_ast(file)
        decorated_nodes, decorator_aliases = get_decorated_nodes(parsed_ast)
        used_keys = {}
        tasks = get_magniv_tasks(
            filepath,
            decorated_nodes,
            decorator_aliases,
            root="tests/unit_tests",
            req="requirements.txt",
            used_keys=used_keys,
        )
        assert tasks[0]["name"] == "hello_world"
        assert tasks[0]["schedule"] == "@hourly"
        assert tasks[0]["location"] == filepath
        assert tasks[0]["requirements_location"] == "requirements.txt"
        assert tasks[0]["description"] == "Hello world, printing the time"
        assert used_keys["goodbye_world_1"] == filepath
        assert used_keys["goodbye_world_2"] == filepath

    def test_get_tasklist(self, file):
        """
        It takes a filepath, parses the file, finds all the decorated functions, and returns a list of
        dictionaries containing the information about each task
        """
        task_list = get_task_list([{"filepath": f"{file}/main.py", "req": None}])
        assert task_list[0]["name"] == "hello_world"
        assert task_list[0]["schedule"] == "@hourly"
        assert task_list[0]["location"] == f"{file}/main.py"
        assert task_list[0]["requirements_location"] is None
        assert task_list[0]["description"] == "Hello world, printing the time"

    def test_get_task_files(self, file):
        """
        `get_task_files` is a function that takes a folder path as an argument and returns a list of
        filepaths to all the files in that folder
        """
        task_files = get_task_files(
            file,
        )
        assert task_files[0]["filepath"] == f"{file}/main.py"

    def test_save(self, file):
        """
        `save_tasks` is a function that takes a folder path as an argument and saves all the tasks in
        :param file: This is the path to the folder where the tasks are stored
        """
        json_pth = f"{file}/dump.json"
        save_tasks(task_folder=file, dump_save_pth=json_pth)
        assert os.path.exists(json_pth) is True

    def test_build(self, file):
        """
        `build` is a function that takes a filepath and a task_folder argument and saves all the tasks
        in that folder to a file

        :param file: a filepath to a Python file
        """
        json_pth = f"{file}/dump.json"
        build(task_folder=file)
        assert os.path.exists(json_pth) is True
        with pytest.raises(OSError):
            build(task_folder="")

    def test_reqs_dependency(self):
        """
        > This function tests that the save_tasks function raises an OSError if the task_folder argument
        is an empty string
        """
        with pytest.raises(OSError):
            save_tasks(task_folder="")

    @pytest.fixture
    def subdir_reqs_file(self, tmpdir):
        """
        It creates a directory structure with a requirements.txt file in it

        :param tmpdir: This is a pytest fixture that creates a temporary directory for us to use
        :return: The path to the requirements.txt file.
        """
        tmpdir.mkdir("tasks").mkdir("subdir").mkdir("subsubdir").join("requirements.txt").write(
            "magniv"
        )
        return str(tmpdir.join("tasks/subdir/subsubdir"))

    def test_subdirreqs_filepath(self, subdir_reqs_file):
        """
        `save_tasks` is a function that takes a folder path as an argument and saves all the tasks in
        that folder to a file

        :param subdir_reqs_file: a filepath to a requirements.txt file in a subdirectory
        """
        json_pth = f"{subdir_reqs_file}/dump.json"
        save_tasks(task_folder=subdir_reqs_file, dump_save_pth=json_pth)
        assert os.path.exists(json_pth) == True

    def test_task_builds_with_valid_custom_resources(self, file):
        task_list = get_task_list([{"filepath": f"{file}/main.py", "req": None}])
        for task in task_list:
            if task["name"] == "resourceful_valid":
                assert task["resources"] == {
                    "limit_cpu": "500mi",
                    "limit_memory": "3Gi",
                }

    def test_task_ignores_invalid_custom_resources(self, file):
        task_list = get_task_list([{"filepath": f"{file}/main.py", "req": None}])
        for task in task_list:
            if task["name"] == "resourceful_invalid":
                assert not task["resources"]

    def test_task_builds_with_webhook_trigger_enabled(self, file):
        task_list = get_task_list([{"filepath": f"{file}/main.py", "req": None}])
        for task in task_list:
            if task["name"] == "playing_webhooky":
                assert task["webhook_trigger"]

    def test_task_builds_without_schedule(self, file):
        task_list = get_task_list([{"filepath": f"{file}/main.py", "req": None}])
        for task in task_list:
            if task["name"] == "task_b":
                assert not task["schedule"]
            elif task["name"] == "c":
                assert not task["schedule"]

    def test_task_builds_with_on_success(self, file):
        task_list = get_task_list([{"filepath": f"{file}/main.py", "req": None}])
        for task in task_list:
            if task["name"] == "task_a":
                assert task["on_success"] == ["task_b", "c"]
            if task["name"] == "task_b":
                assert task["is_called_by"] == ["task_a", "c"]
            if task["name"] == "c":
                assert task["is_called_by"] == ["task_a"]
                assert task["webhook_trigger"]



import hashlib
import importlib.util
import os
import sys


def run(filepath, function_name):
    """
    It imports a Python file and runs a function from it

    :param filepath: The path to the file containing the function to be run
    :param function_name: The name of the function to run
    :return: The return value of the function.
    """
    org_mod_name, _ = os.path.splitext(os.path.split(filepath)[-1])
    path_hash = hashlib.sha1(filepath.encode("utf-8")).hexdigest()
    mod_name = f"unusual_prefix_{path_hash}_{org_mod_name}"
    spec = importlib.util.spec_from_file_location(mod_name, filepath)
    mod = importlib.util.module_from_spec(spec)
    # Hack to fix project imports by adding project directory to Python path
    sys.path.append(f"{os.getcwd()}/tasks")
    spec.loader.exec_module(mod)
    func = getattr(mod, function_name)
    return func()



import ast
import re
from functools import update_wrapper
from typing import Callable, Dict

from croniter import CroniterBadCronError, CroniterBadDateError, CroniterNotAlphaError, croniter

task_kwargs = {
    "on_success": {"ast_type": ast.List, "readable": "list[str]"},
    "resources": {"ast_type": ast.Dict, "readable": "dict"},
    "key": {"ast_type": ast.Constant, "readable": "str"},
    "schedule": {"ast_type": ast.Constant, "readable": "str"},
    "webhook_trigger": {"ast_type": ast.Constant, "readable": "str"},
    "description": {"ast_type": ast.Constant, "readable": "str"},
}


class Task:
    """

    This class contains all the relevant information for each task, it is used as a decorator placed around functions
    that are to be deployed.

    :param function: the function that is being decorated
    :param schedule: the cron-style schedule interval that determines when this function runs
    :param webhook_trigger: boolean flag to enable task being triggered through webhook
    :param resources: the cpu and memory requirements for this function
    :param description: description of the function, to be used for the auto generated documentation
    :param key: the unique key that will reference the function, default is the function of the name
    :param on_success: list of downstream task keys that are triggered after this task's successful completion. Adding multiple tasks to this list will add multiple downstream calls.
    """

    KEY_PATTERN = r"^[\w\-.]+$"

    def __init__(
        self,
        function,
        schedule=None,
        webhook_trigger=False,
        resources=None,
        description=None,
        key=None,
        on_success=None,
    ) -> None:
        if schedule is not None and not self._is_valid_schedule(schedule):
            raise ValueError(f"{schedule} is not a valid cron schedule")
        self.schedule = schedule
        self.webhook_trigger = webhook_trigger
        self.resources = self._make_valid_resources(resources)
        self.description = description
        self.function = function
        self.name = function.__name__
        self.key = key if key else self.name
        if key and not self._is_valid_key(key):
            raise ValueError(
                f"{key} is not a valid key, the key can only contain alphanumeric characters, -, _, . and space."
            )
        self.on_success = on_success if on_success else []

        update_wrapper(self, function)

    def as_dict(self) -> dict:
        return {
            "schedule": self.schedule,
            "webhook_trigger": self.webhook_trigger,
            "resources": self.resources,
            "description": self.description,
            "name": self.name,
            "key": self.key,
            "on_success": self.on_success,
        }

    def __call__(self, *args, **kwds) -> Callable:
        return self.function(*args, **kwds)

    def _is_valid_schedule(self, schedule) -> bool:
        """
        It returns `True` if the given `schedule` is a valid cron expression, and `False` otherwise
        Based on whatever Airflow accepts as a valid cron expression

        :param schedule: The cron schedule for the task
        :return: A boolean value.
        """
        try:
            croniter(schedule)
            return True
        except (
            CroniterBadCronError,
            CroniterBadDateError,
            CroniterNotAlphaError,
        ) as cron_e:
            return False
        return bool(re.match(Task.CRON_PATTERN, schedule))

    def _is_valid_key(self, key) -> bool:
        """
        It returns `True` if the given `key` matches the regular expression `Task.KEY_PATTERN`, and `False`
        otherwise

        :param key: The key of the task
        :return: A boolean value.
        """
        return bool(re.match(Task.KEY_PATTERN, key))

    def _make_valid_resources(self, resources) -> Dict[str, str]:
        """
        Takes in a resources dictionary with generic parameters and converts it to the appropriate syntax
        for our infrastructure
        #TODO: enforce reasonable limits for quantity of resources
        :param key: The generic resources dict from the task decorator
        :return: A dict with correct syntax
        """
        clean_dict = {}
        if resources is None:
            return resources
        valid_resources = set(["cpu", "memory"])
        invalid_resources = set([r.lower() for r in resources.keys()]) - valid_resources
        if len(invalid_resources) > 0:
            raise KeyError(
                f"{invalid_resources} are not valid resources for a task. Valid resources are {valid_resources}"
            )
        if "cpu" in resources.keys():
            clean_dict["limit_cpu"] = resources["cpu"]
        if "memory" in resources.keys():
            clean_dict["limit_memory"] = resources["memory"]
        return clean_dict


def task(
    _func=None,
    *,
    schedule=None,
    webhook_trigger=False,
    resources=None,
    description=None,
    key=None,
    on_success=None,
) -> Callable:
    """
    If they pass in a function, then we raise an error. If they dont pass in a function, then we return
    a wrapper function that takes a function as an argument

    :param _func: This is the function that is being wrapped
    :param schedule: This is the schedule that the task will run on. It can be a cron string, or a
    datetime.timedelta object, or None
    :param webhook_trigger: Specifices whether this task can be triggered via webhook (see dashboard for webhook URL)
    :param resources: The cpu and memory requirements for this function
    :param description: A description of the task
    :param key: This is the name of the task key. It is used to identify the task in the database
    :param on_success: This is a list of downstream task keys (usually the @task-decorated-function's name) that
    are triggered by this task on successful completion. Adding multiple tasks to this list will add multiple downstream triggers. (list[str])
    :return: A function that takes in a function and returns a task instance.
    """
    if _func is not None:  # this means they did not pass in any arguments like @magniv
        # The reason we do this here is bc in the case they dont pass arguments we dont need the extra wrapper below.
        raise ValueError("You must use arguments with magniv, it can not be called alone")

    def wrapper(function):
        return Task(
            function,
            schedule=schedule,
            webhook_trigger=webhook_trigger,
            resources=resources,
            description=description,
            key=key,
            on_success=on_success,
        )

    return wrapper



# Traverse all python files in this directory and one child lower as well.
import ast
import os
import platform
import sys
from typing import Dict, List, NoReturn, Tuple

from magniv.core import Task, task_kwargs
from magniv.utils.utils import _save_to_json


def _get_python_version(root):
    """
    It returns the version of Python that is running

    :param root: The root directory of the project
    :return: The version of Python that is being used.
    """
    return platform.python_version()


def _get_owner(root):
    """
    It returns the string "local"

    :param root: The root directory of the project
    :return: The owner of the root directory.
    """
    return "local"


def _get_ast_alias(names: List, key: str):
    """
    It takes a list of names and a key, and returns the alias of the name in the list that matches the
    key, or None if there is no match

    Args:
      names (List): List[ast.alias]
      key (str): The name of the variable we're looking for.

    Returns:
      The alias of the name, or the name itself if no alias is present.
    """
    matching_name = next((i for i in names if i.name == key), None)
    if not matching_name:
        return None
    return matching_name.asname or matching_name.name


def _get_decorator_name(func):
    """
    It takes a Python AST node and recursively returns the decorator name.

    Args:
      func: The function being decorated.

    Returns:
      The name of the decorator.
    """
    if isinstance(func, ast.Name) and hasattr(func, "id"):
        return func.id
    if (
        isinstance(func, ast.Attribute)
        and hasattr(func, "value")
        and hasattr(func, "attr")
    ):
        return ".".join([_get_decorator_name(func.value), func.attr])
    return ""

def _extract_task_kwarg_value(kw: ast.keyword, func_name: str):
    """
    Extract the kwarg value from the given @task kwarg AST object. Raise error if invalid kwarg.

    Returns:
      Parsed AST value given to kw.arg
    """
    if kw.arg not in task_kwargs:
        raise ValueError(f'Unknown kwarg "{kw.arg}" provided to task {func_name}')

    arg_type = task_kwargs[kw.arg]["readable"]
    arg_type_ast = task_kwargs[kw.arg]["ast_type"]
    if not isinstance(kw.value, arg_type_ast):
        raise ValueError(f'Keyword argument "{kw.arg}" to task {func_name} is required to be a {arg_type}')

    if arg_type_ast == ast.List:
        return [i.value for i in kw.value.elts]
    if arg_type_ast == ast.Dict:
        return dict(
            zip(
                [key.value for key in kw.value.keys],
                [val.value for val in kw.value.values],
            )
        )
    if arg_type_ast == ast.Constant:
        return kw.value.value

def get_decorated_nodes(parsed_ast: ast.AST) -> List:
    """
    It returns decorator nodes and decorator aliases.

    Args:
      parsed_ast (ast.AST): The parsed AST of the file.

    Returns:
      A list of nodes that are either a FunctionDef or AsyncFunctionDef and have a decorator_list.
      A list of aliases that can be used for Magniv task decorators.
    """

    decorator_aliases = []
    decorated_nodes = []
    for node in ast.walk(parsed_ast):
        # Add decorated functions
        if (
            isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
            and len(node.decorator_list) > 0
        ):
            decorated_nodes.append(node)
            continue

        # Add possible decorator aliases
        if isinstance(node, ast.Import):
            import_variants = [
                # import magniv --- @magniv.core.task
                {"import": "magniv", "decorator_suffix": ".core.task"},
                # import magniv.core --- @magniv.core.task
                {"import": "magniv.core", "decorator_suffix": ".task"},
            ]
            for variant in import_variants:
                if alias := _get_ast_alias(node.names, variant["import"]):
                    decorator_aliases.append(alias + variant["decorator_suffix"])
        elif isinstance(node, ast.ImportFrom):
            if node.module == "magniv.core":
                # from magniv.core import task --- @task
                if alias := _get_ast_alias(node.names, "task"):
                    decorator_aliases.append(alias)
            elif node.module == "magniv":
                # from magniv import core --- @core.task
                if alias := _get_ast_alias(node.names, "core"):
                    decorator_aliases.append(f"{alias}.task")
    return decorated_nodes, decorator_aliases


def get_magniv_tasks(
    filepath: str,
    decorated_nodes: List,
    decorator_aliases: List,
    root: str,
    req: str,
    used_keys: Dict,
) -> Tuple[List, Dict]:
    """
    It returns all Magniv decorated tasks from a list of python functions that are decorated

    Args:
      filepath (str): The path to the file that contains the tasks.
      decorated_nodes (List): A list of all the nodes in the graph.
      root (str): The root directory of the project.
      req (str): The requirement file that is being parsed.
      used_keys (Dict): A dictionary of keys that have been used in the file.
    """
    tasks = []
    for node in decorated_nodes:
        core_values = {
            "location": filepath,
            "python_version": _get_python_version(root),
            "owner": _get_owner(root),
            "requirements_location": req,
            "line_number": node.lineno,
            "description": None,
        }
        for decorator in node.decorator_list:
            if hasattr(decorator, "func"):
                decorator_name = _get_decorator_name(decorator.func)
                if decorator_name not in decorator_aliases:
                    continue

                constructed_decorator_values = {}
                for kw in decorator.keywords:
                    constructed_decorator_values[kw.arg] = _extract_task_kwarg_value(kw, node.name)
                try:
                    dummy_function = lambda x: None
                    dummy_function.__name__ = node.name
                    task = Task(dummy_function, **constructed_decorator_values)
                    decorator_values = task.as_dict()
                except Exception as e:
                    raise Exception(
                        f"Building the task {node.name} in {core_values['location']} on line {core_values['line_number']} raised the following {type(e).__name__}: {e}"
                    )

                info = {**core_values, **decorator_values}
                if missing_reqs := list({"schedule"} - set(info)):
                    raise ValueError(
                        "Task missing required variables, please resolve by defining ("
                        + ",".join([f" {x} " for x in missing_reqs])
                        + f') in the task decorator for function {info["name"]} at {info["location"]} line {info["line_number"]}'
                    )
                if info["key"] in used_keys:
                    raise ValueError(
                        f'Task "{info["key"]}" in file {filepath} is using "{info["key"]}" as a key which is already used in {used_keys[info["key"]]}, please resolve by changing one of the keys'
                    )
                used_keys[info["key"]] = filepath
                tasks.append(info)

    return tasks


def get_task_files(task_folder: str) -> List:
    """
    It gets all the files in a task folder

    Args:
      task_folder (str): the folder where the task is located

    Returns:
      A list of dictionaries, each dictionary contains the filepath and the requirements file for that
    file.
    """
    root_req = f"{task_folder}/requirements.txt"
    if not os.path.exists(root_req):
        raise OSError(f"{root_req} not found")
    task_files = []
    for root, dirs, files in os.walk(task_folder):
        req = (
            f"{root}/requirements.txt"
            if os.path.exists(f"{root}/requirements.txt")
            else root_req
        )
        if req is None:
            raise OSError(
                f'requirements.txt not found for path "{root}", either add one to this directory or the root directory'
            )
        for file_name in files:
            if file_name.endswith(".py"):
                fileinfo = {"filepath": f"{root}/{file_name}", "req": req}
                task_files.append(fileinfo)
    return task_files


def get_task_list(
    task_files: List,
    task_folder: str = None,
    tasks_list: List = None,
    used_keys: Dict = None,
) -> List:
    """
    "Get a list of tasks from a list of task files."

    Args:
      task_files (List): List of task files to be processed.
      task_folder (str): The folder where the tasks are located.
      tasks_list (List): This is the list of tasks that will be returned.
      used_keys (Dict): A dictionary of keys that have already been used.
    """
    if tasks_list is None:
        tasks_list = []
    if used_keys is None:
        used_keys = {}
    for fileinfo in task_files:
        with open(fileinfo["filepath"]) as f:
            try:
                parsed_ast = ast.parse(f.read())
            except UnicodeDecodeError as e:
                continue
            decorated_nodes, decorator_aliases = get_decorated_nodes(parsed_ast)
            tasks_list.extend(
                get_magniv_tasks(
                    fileinfo["filepath"],
                    decorated_nodes,
                    decorator_aliases,
                    root=task_folder,
                    req=fileinfo["req"],
                    used_keys=used_keys,
                )
            )

    # Check for invalid calls
    for task in tasks_list:
        for triggered_task in task["on_success"]:
            if triggered_task not in used_keys:
                raise ValueError(
                    f'Task "{task["key"]}" triggers unkown task "{triggered_task}".')

    # Add is_called_by to each task
    called_by = {key: [] for key in used_keys}
    for task in tasks_list:
        for triggered_task in task["on_success"]:
            called_by[triggered_task].append(task["key"])
    for task in tasks_list:
        task['is_called_by'] = called_by[task["key"]] 

    return tasks_list


def save_tasks(
    task_folder: str,
    save_dir: bool = False,
    dump_save_pth: str = "./dump.json",
    tasks_list: List = None,
) -> NoReturn:
    """
    `save_tasks` saves a list of tasks to a json file

    Args:
      task_folder (str): The folder where the tasks are located.
      save_dir (bool): If True, the directory of the task will be saved. Defaults to False
      dump_save_pth (str): The path to the file where the dump will be saved. Defaults to ./dump.json
      tasks_list (List): List = None
    """
    task_files = get_task_files(task_folder)
    tasks_list = get_task_list(task_files, task_folder)
    if save_dir:
        dump_save_pth = task_folder + dump_save_pth[1:]
    _save_to_json(tasks_list, filepath=dump_save_pth)


def build(task_folder: str = None):
    """
    We walk through the `tasks` directory, and for each file we find, we check if it's a Python file,
    and if it is, we get the tasks from it and add them to our list of tasks

    Args:
      task_folder (str): The folder that contains all of your tasks files. If you don't specify this, it
    will default to `./tasks`
    """
    save_dir = False
    if task_folder is None:
        task_folder = "./tasks"
    else:
        save_dir = True
    if not os.path.exists(task_folder):
        raise OSError(
            "You must have a tasks folder that contains all of your tasks files"
        )

    # Hack to fix project imports by adding project directory to Python path
    sys.path.append(os.getcwd() + task_folder)
    save_tasks(task_folder, save_dir=save_dir)



import fileinput
import os
import shutil

import docker
from dotenv import dotenv_values

from magniv.utils.utils import _create_cloud_build


def export_to_airflow(
    task_list,
    gcp=False,
    gcp_project_id=None,
    gcp_dag_folder=None,
    callback_hook=None,
    kubernetes_startup_timeout=None,
    env_file_path=None,
):
    """
    It takes a list of tasks, creates a docker image for each task, creates a dag file for each task,
    and then creates a cloud build file to build the docker images.

    :param task_list: a list of dictionaries, each dictionary containing the following keys:
    :param gcp: If you want to use Google Cloud Platform, set this to True, defaults to False (optional)
    :param gcp_project_id: The project id of your GCP project
    :param gcp_dag_folder: The folder where the dag files will be stored
    :param callback_hook: This is the URL of the callback function that will be called when the task is
    completed
    :param kubernetes_startup_timeout: This is the time in seconds that Airflow will wait for the
    Kubernetes pod to start
    :param env_file_path: This is the path to the environment file that you want to use
    """
    dag_template_filename = "dag-template.py"
    dag_template_directory = f"{os.path.dirname(__file__)}/{dag_template_filename}"
    docker_image_info = []
    for task_info in task_list:
        print("starting task .... ")
        new_filename = f'dags/{task_info["owner"]}/{task_info["key"]}.py'
        print(new_filename)
        if not os.path.exists("dags/"):
            os.mkdir("dags")
        if not os.path.exists(f'dags/{task_info["owner"]}/'):
            os.mkdir(f'dags/{task_info["owner"]}/')
        shutil.copyfile(dag_template_directory, new_filename)
        print("creating docker image ... ")
        docker_name, path = _create_docker_image(
            task_info["python_version"],
            task_info["requirements_location"],
            task_info["key"],
            gcp=gcp,
            gcp_project_id=gcp_project_id,
            env_file_path=env_file_path,
        )
        docker_name, path = (None, None)
        print("docker image created!")
        docker_image_info.append((docker_name, path))
        print("creating dag ... ")
        with fileinput.input(new_filename, inplace=True) as f:
            for line in f:
                schedule = (
                    "None" if task_info["schedule"] is None else f"""'{task_info["schedule"]}'"""
                )
                line = (
                    line.replace("dag_id", f"""'{task_info["key"]}'""")
                    .replace("ownertoreplace", f"""'{task_info["owner"]}'""")
                    .replace("scheduletoreplace", schedule)
                    .replace("imagetoreplace", f"'{docker_name}'")
                    .replace("filetoreplace", task_info["location"])
                    .replace("functiontoreplace", task_info["name"])
                    .replace(
                        "resourcesdicttoreplace",
                        str(task_info["resources"])
                        if "resources" in task_info.keys()
                        else "None",
                    )
                    .replace(
                        "callbackhooktoreplace",
                        f"'{callback_hook}'" if callback_hook is not None else "None",
                    )
                    .replace(
                        "successtoreplace", "_on_success" if callback_hook is not None else "None"
                    )
                    .replace(
                        "failuretoreplace", "_on_failure" if callback_hook is not None else "None"
                    )
                    .replace("startuptoreplace", str(kubernetes_startup_timeout))
                )

                print(line, end="")
        print("dag created!")
    if gcp:
        _create_cloud_build(docker_image_info, gcp_dag_folder)


def _create_docker_image(
    python_version,
    requirements,
    key,
    gcp=False,
    gcp_project_id=None,
    env_file_path=None,
):
    """
    It creates a Dockerfile in the same directory as the requirements.txt file, and then builds a docker
    image from that Dockerfile

    :param python_version: The version of python you want to use
    :param requirements: The path to the requirements.txt file
    :param key: This is the name of the function key. It's used to name the docker image
    :param gcp: If you're using GCP, set this to True, defaults to False (optional)
    :param gcp_project_id: The ID of the GCP project you want to use
    :param env_file_path: The path to the .env file that contains the environment variables that you
    want to pass to the Docker container
    :return: The docker image name and the path to the dockerfile
    """
    path = "/".join(requirements.split("/")[:-1])
    if not gcp:
        requirements = "requirements.txt"
    environment_arguments = ""
    if env_file_path is not None:
        env_values_dict = dotenv_values(env_file_path)
        environment_arguments = "\n".join(
            [f'ENV {key}="{env_values_dict[key]}"' for key in env_values_dict]
        )
    dockerfile = f"""
# syntax=docker/dockerfile:1

FROM python:{python_version}
COPY {requirements} requirements.txt
{environment_arguments}
RUN pip3 install -r requirements.txt

COPY . .
                """
    with open(f"{path}/Dockerfile", "w") as fo:
        fo.write(dockerfile)
    docker_name = f"{key}dockerimage"
    if gcp:
        docker_name = f"gcr.io/{gcp_project_id}/{docker_name}"
    else:
        client = docker.from_env()
        client.images.build(path=path, tag=docker_name)
    return docker_name, path



import pytest

from magniv.build.build import get_decorated_nodes
from tests.unit_tests.test_build import TestBuild

TEST_FILE = """from datetime import datetime

import magniv
from magniv.core import task
from magniv import core as mc


@magniv.core.task(schedule="@hourly", description="Hello world, printing the time", key="goodbye_world_1")
def hello_world():
    print(f"Hello world, the time is {datetime.now()}")

@task(schedule="@daily", description="Hello world, printing a second time", key="goodbye_world_2")
def hello_world_2():
    print(f"Hello world, the time is {datetime.now()}")

@mc.task(schedule="@daily", description="Hello world, printing a third time", key="goodbye_world_3")
def hello_world_3():
    print(f"Hello world, the time is {datetime.now()}")


def dummy_task():
    print("This is a dummy task")
    return True

if __name__ == "__main__":
    hello_world()
"""


class TestBuild(TestBuild):
    @pytest.fixture
    def file(self, tmpdir):
        """
        It creates a directory called tasks, creates a file called main.py in that directory, and writes
        the contents of the TEST_FILE variable to that file. It then creates a file called
        requirements.txt in the tasks directory and writes the word magniv to it. Finally, it returns
        the path to the tasks directory

        Args:
          tmpdir: This is a pytest fixture that creates a temporary directory for us to use.

        Returns:
          The path to the directory where the file is located.
        """
        tmpdir.mkdir("tasks").join("main.py").write(TEST_FILE)
        tmpdir.join("tasks/requirements.txt").write("magniv")
        return str(tmpdir.join("tasks"))

    def _extracted_from_test_get_decorated_nodes(self, decorated_nodes, arg1, arg2, arg3, arg4):
        """
        A function that checks if the name of the node is the same as the name of the function, and if
        the decorator list has the same keywords as the function.

        Args:
          decorated_nodes: the list of nodes that have been decorated
          arg1: The index of the decorated node in the list of decorated nodes.
          arg2: the name of the function
          arg3: the schedule
          arg4: the name of the function
        """
        assert decorated_nodes[arg1].name == arg2
        assert decorated_nodes[arg1].decorator_list[0].keywords[0].arg == "schedule"
        assert decorated_nodes[arg1].decorator_list[0].keywords[0].value.s == arg3
        assert decorated_nodes[arg1].decorator_list[0].keywords[2].value.s == arg4

    def test_get_decorated_nodes(self, file):
        """
        `get_decorated_nodes` returns a list of all the decorated functions in a Python file
        """
        parsed_ast, filepath = self.get_ast(file)
        decorated_nodes, _ = get_decorated_nodes(parsed_ast)
        self._extracted_from_test_get_decorated_nodes(
            decorated_nodes, 0, "hello_world", "@hourly", "goodbye_world_1"
        )
        self._extracted_from_test_get_decorated_nodes(
            decorated_nodes, 1, "hello_world_2", "@daily", "goodbye_world_2"
        )
        self._extracted_from_test_get_decorated_nodes(
            decorated_nodes, 2, "hello_world_3", "@daily", "goodbye_world_3"
        )



import json


def _save_to_json(obj, filepath):
    """
    It takes a Python object and a filepath, and saves the object to the filepath as a JSON file

    :param obj: the object to be saved
    :param filepath: The path to the file to save the data to
    """
    with open(filepath, "w") as fo:
        json.dump(obj, fo)


def _get_tasks_json(filepath):
    """
    It opens the file at the given filepath, reads the contents of the file, and returns the contents as
    a Python object

    :param filepath: The path to the JSON file containing the tasks
    :return: A list of dictionaries.
    """
    with open(filepath) as fo:
        task_list = json.load(fo)
    return task_list


def _create_cloud_build(docker_image_info, gcp_dag_folder):
    """
    It creates a cloudbuild.yaml file that will build all the docker images and then sync the dags
    folder to the GCP bucket

    :param docker_image_info: A list of tuples. Each tuple contains the name of the image you want to
    build and the path to the Dockerfile
    :param gcp_dag_folder: The GCP bucket where the DAGs will be stored
    """
    with open("./cloudbuild.yaml", "w") as fo:
        fo.write(
            f"steps:\n- name: gcr.io/cloud-builders/gsutil\n  args:\n    - '-m'\n    - 'rsync'\n    - '-d'\n    - '-r'\n    - 'dags'\n    - '{gcp_dag_folder}'"
        )
        gcp_image_names = []
        for docker_info in docker_image_info:
            gcp_image_name = docker_info[0]
            path = docker_info[1]
            gcp_image_names.append(gcp_image_name)
            fo.write(
                f"\n- name: 'gcr.io/cloud-builders/docker'\n  args: [ 'build', '-t','{gcp_image_name}', '-f', '{path}/Dockerfile', '.' ]"
            )
        fo.write("\nimages: [{}]".format(",".join(f"'{image}'" for image in gcp_image_names)))



import ast
import os
from typing import Union

import pytest

from magniv.build.build import (
    build,
    get_decorated_nodes,
    get_magniv_tasks,
    get_task_files,
    get_task_list,
    save_tasks,
)

TEST_FILE = """from datetime import datetime

from magniv.core import task

trigger_lst = ['task_b']

@task(schedule="@daily", _on_success=trigger_lst)
def task_a():
    print("I")

@task()
def task_b():
    print("i'm only triggered by task A")


if __name__ == "__main__":
    hello_world()
"""


class TestBuild:
    @pytest.fixture
    def file(self, tmpdir):
        tmpdir.mkdir("tasks").join("main.py").write(TEST_FILE)
        tmpdir.join("tasks/requirements.txt").write("magniv")
        return str(tmpdir.join("tasks"))

    def test_build(self, file):
        """
        `build` is a function that takes a filepath and a task_folder argument and saves all the tasks
        in that folder to a file

        :param file: a filepath to a Python file
        """
        json_pth = f"{file}/dump.json"
        with pytest.raises(ValueError):
            build(task_folder=file)



from magniv.export.airflow import export_to_airflow
from magniv.utils.utils import _get_tasks_json


def export(
    gcp=False,
    gcp_project_id=None,
    gcp_dag_folder=None,
    callback_hook=None,
    kubernetes_startup_timeout=None,
    env_file_path=None,
):
    """
    It takes a list of tasks and exports them to Airflow

    :param gcp: If you want to export to GCP, set this to True, defaults to False (optional)
    :param gcp_project_id: The GCP project ID where the DAGs will be deployed
    :param gcp_dag_folder: The folder in which the DAGs will be created
    :param callback_hook: A function that will be called after the DAG is exported
    :param kubernetes_startup_timeout: The amount of time to wait for the Kubernetes pod to start up
    :param env_file_path: The path to the environment file
    """
    task_list = _get_tasks_json("./dump.json")
    export_to_airflow(
        task_list,
        gcp=gcp,
        gcp_project_id=gcp_project_id,
        gcp_dag_folder=gcp_dag_folder,
        callback_hook=callback_hook,
        kubernetes_startup_timeout=kubernetes_startup_timeout,
        env_file_path=env_file_path,
    )



from setuptools import find_packages, setup

setup(
    name="magniv",
    description="Magniv Core Library",
    author_email="hello@magniv.io",
    url="https://www.magniv.io",
    project_urls={
        "Documentation": "https://docs.magniv.io",
    },
    version="0.1.55",
    py_modules=["magniv"],
    packages=find_packages(),
    install_requires=["Click", "docker", "python-dotenv", "requests", "croniter"],
    entry_points={
        "console_scripts": [
            "magniv-cli = magniv.scripts.magniv:cli",
        ],
    },
    include_package_data=True,
)



import click

from magniv.build import build as m_build
from magniv.export import export as m_export
from magniv.run import run as m_run


@click.group()
def cli():
    pass


@cli.command()
def build():
    return m_build()


@cli.command()
@click.option("--gcp", is_flag=True)  # GCP creeates the dockerfiles and cloubuild.yaml
@click.option("--gcp-project-id")
@click.option("--gcp-dag-folder")
@click.option("--callback-hook")
@click.option("--kubernetes-startup-timeout", default=120)
@click.option("--env-file-path")
def export(
    gcp,
    gcp_project_id,
    gcp_dag_folder,
    callback_hook,
    kubernetes_startup_timeout,
    env_file_path,
):
    return m_export(
        gcp=gcp,
        gcp_project_id=gcp_project_id,
        gcp_dag_folder=gcp_dag_folder,
        callback_hook=callback_hook,
        kubernetes_startup_timeout=kubernetes_startup_timeout,
        env_file_path=env_file_path,
    )


@cli.command()
@click.argument("filepath")
@click.argument("function_name")
def run(filepath, function_name):
    return m_run(filepath, function_name)
